/**
 * SWE KEx-DatexII (de.bsvrz.kex.datex2)
 * 
 * Copyright (C) 2014 TraffGo Road GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * TraffGo Road GmbH
 * Ostwall 11-19
 * 47798 Krefeld
 * Phone: +49-2151-3295058
 * mailto: info@traffgoroad.de
 */

package de.bsvrz.kex.datex2.allgemein.impl;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.NoSuchElementException;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.ReferenceArray;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ClientApplication;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.kex.datex2.allgemein.ErweiterterLogger;
import de.bsvrz.kex.datex2.allgemein.KonfigurationsModul;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Implementierung des Interface ErweiterterLogger; Daten werden ber den
 * Datenverteiler versendet (Statusmeldungen durch spezielle Attributgruppe,
 * Meldungen an Operator durch Betriebsmeldungen).
 * 
 * @author TraffGo Road GmbH
 * @version $Id$
 */
public class ErweiterterLoggerDaV implements ClientSenderInterface, ErweiterterLogger {
  /** Aspekt zum Senden von Betriebsmeldungen. */
  private static final String ASP_MELDUNG = "asp.information";
  
  /** Attributgruppe zum Senden von Betriebsmeldungen. */
  private static final String ATG_MELDUNG = "atg.betriebsMeldung";
  
  /** Schlssel zur Angabe des Betriebsmeldungsobjekts. */
  private static final String KEY_BETRIEBSMELDUNGSOBJEKT = "BetriebsmeldungsObjekt";
  
  /** Schlssel zur Angabe des Statusmeldungsobjekts. */
  private static final String KEY_STATUSOBJEKT = "StatusObjekt";
  
  /** Aspekt zum Senden von Statusmeldungen. */
  private static final String ASP_STATUSMELDUNG = "asp.zustand";
  
  /** Attributgruppe zum Senden von Statusmeldungen. */
  private static final String ATG_STATUSMELDUNG = "atg.statusmeldung";
  
  /** Vermeidung doppelter Meldungen: Erneutes Senden der Meldung (in Stunden). */
  private static final int ERNEUTES_SENDEN_IN_STUNDEN = 24;
  
  /** Maximale Wartezeit auf positive Sendesteuerung (in Millisekunden). */
  private static final long MAX_WAIT = 5000;
  
  /** Applikations-Kennung fr Betriebsmeldungen. */
  private static final String APPLIKATIONS_KENNUNG = "KEx-DatexII";
  
  /** Datenverteiler-Verbindung. */
  private ClientDavInterface davConnection;
  
  /** Prfix fr Log-Ausgaben. */
  private String logPrefix = "";
  
  /** Datenverteiler-Objekt zum Senden von Statusmeldungen. */
  private SystemObject statusObjekt;

  /** Datenverteiler-Objekt zum Senden von Betriebsmeldungen. */
  private SystemObject betriebsmeldungsObjekt;

  /** DataDescription zum Senden von Betriebsmeldungen. */
  private DataDescription betriebsmeldungDD;
  
  /** Logging. */
  private static final Debug LOGGER = Debug.getLogger();
  
  /** Letzter gesendeter Text einer Betriebsmeldung. */
  private String letzteMeldungAnOperator = "";
  
  /** Zeitpunkt der letzten erfolgreichen bermittlung einer Betriebsmeldung. */
  private Calendar letzteMeldungAnOperatorZeitpunkt = Calendar.getInstance();
  
  /** Datenverteiler-Applikations-ID. */
  private long applicationID;
  
  /** Laufende Nummer der Betriebsmeldungen (auf Applikation bezogen!). */
  private static Integer lfdNummer = Integer.valueOf(0);
  
  /** Lock. */
  private static Object lockLfdNummer = new Object();
  
  /** Typ der Applikation. */
  private SystemObjectType applicationType;
  
  /** Name der Plugin-Klasse. */
  private String pluginClassname;
  
  /**
   * Konstruktor. Anmeldung als Sender fr Status- und Betriebsmeldungen.
   * 
   * @param config Referenz auf Konfigurationsmodul
   * @param davConnection Referenz auf Datenverteiler-Verbindung
   */
  public ErweiterterLoggerDaV(final KonfigurationsModul config, final ClientDavInterface davConnection) {
    this.davConnection = davConnection;
    this.logPrefix = "In Kontext '" + config.getContext() + "': ";
    
    pluginClassname = config.getString("Plug-in-Klasse");

    if (config.hasData(KEY_STATUSOBJEKT)) {
      String statusPid = config.getString(KEY_STATUSOBJEKT);
      
      this.statusObjekt = davConnection.getDataModel().getObject(statusPid);
      if (statusObjekt == null) {
        LOGGER.error(logPrefix + "Statusmeldungen nicht aktivierbar: Objekt '" + statusPid + "' nicht gefunden.");
      } else {
        DataDescription ddSend1 = new DataDescription(davConnection.getDataModel().getAttributeGroup(ATG_STATUSMELDUNG), 
            davConnection.getDataModel().getAspect(ASP_STATUSMELDUNG));
        try {
          davConnection.subscribeSender(this, statusObjekt,  ddSend1,  SenderRole.source());
        } catch (Exception e) {
          statusObjekt = null;
          LOGGER.error(logPrefix + "Kann Statusmeldung nicht senden, da Initialisierung fehlerhaft.", e);
        }
      }
    } else {
      LOGGER.warning(logPrefix + "Kein Statusmeldungsobjekt angegeben, Statusmeldungen werden nicht ausgegeben.");
    }

    if (config.hasData(KEY_BETRIEBSMELDUNGSOBJEKT)) {
      String betriebsmeldungsPid = config.getString(KEY_BETRIEBSMELDUNGSOBJEKT);
      this.betriebsmeldungsObjekt = davConnection.getDataModel().getObject(betriebsmeldungsPid);
      if (betriebsmeldungsObjekt == null) {
        LOGGER.error(logPrefix + "Betriebsmeldungen nicht aktivierbar: Objekt '" + betriebsmeldungsPid 
            + "' nicht gefunden.");        
      } else {
        betriebsmeldungDD = new DataDescription(davConnection.getDataModel().getAttributeGroup(ATG_MELDUNG), 
            davConnection.getDataModel().getAspect(ASP_MELDUNG));
        try {
          davConnection.subscribeSender(this, betriebsmeldungsObjekt,  betriebsmeldungDD,  SenderRole.sender());
        } catch (Exception e) {
          betriebsmeldungsObjekt = null;
          LOGGER.error(logPrefix + "Kann Betriebsmeldung nicht senden, da Initialisierung fehlerhaft.", e);
        }

        ClientApplication application = davConnection.getLocalApplicationObject();

        applicationID = application.getId();
        applicationType = application.getType();        
      }
    } else {
      LOGGER.warning(logPrefix + "Kein Betriebsmeldungsobjekt angegeben, Betriebsmeldungen werden nicht ausgegeben.");
    }
  }
  
  @Override
  public void sendeAnOperator(final String msg) {
    if (betriebsmeldungsObjekt != null) {
      if (letzteMeldungAnOperator.equals(msg)) {
        Calendar c = Calendar.getInstance();
        c.add(Calendar.HOUR_OF_DAY, -ERNEUTES_SENDEN_IN_STUNDEN);
        if (c.before(letzteMeldungAnOperatorZeitpunkt)) {
          return;
        }
      }
      final Data data = davConnection.createData(davConnection.getDataModel().getAttributeGroup(ATG_MELDUNG));

      data.getUnscaledValue("ApplikationsID").set(applicationID);
      synchronized (lockLfdNummer) {
        data.getUnscaledValue("LaufendeNummer").set(lfdNummer);
        lfdNummer++;
      }
      data.getReferenceValue("ApplikationsTyp").setSystemObject(applicationType);
      if (statusObjekt != null) {
        ReferenceArray refArray = data.getReferenceArray("Referenz");
        refArray.set(statusObjekt);
      }
      data.getTextValue("ApplikationsKennung").setText(APPLIKATIONS_KENNUNG);

      SimpleDateFormat sdf = new SimpleDateFormat("-yyyy-MM-dd");
      data.getTextValue("ID").setText(pluginClassname + sdf.format(new Date()));
      data.getTextValue("MeldungsTyp").setText("System");
      data.getTextValue("MeldungsTypZusatz").setText("");
      data.getTextValue("MeldungsKlasse").setText("Fehler");
      data.getTextValue("Status").setText("Meldung");

      final Data urlasser = data.getItem("Urlasser");
      urlasser.getReferenceValue("BenutzerReferenz").setSystemObject(davConnection.getLocalUser());
      urlasser.getTextValue("Ursache").setText("");
      urlasser.getTextValue("Veranlasser").setText("");

      data.getTextValue("MeldungsText").setText(msg);

      ResultData resultData = new ResultData(betriebsmeldungsObjekt, betriebsmeldungDD, davConnection.getTime(), data);   
      try {
        davConnection.sendData(resultData);
        letzteMeldungAnOperator = msg;
      } catch (DataNotSubscribedException | SendSubscriptionNotConfirmed e) {
        LOGGER.error(logPrefix + "Betriebsmeldung konnte nicht versendet werden: ", e);
      }
    }
  }
  
  @Override
  public void dataRequest(final SystemObject arg0, final DataDescription arg1, final byte arg2) {   
  }

  @Override
  public boolean isRequestSupported(final SystemObject arg0, final DataDescription arg1) {
    return false;
  }

  @Override
  public void sendeStatus(final String msg) {
    if (statusObjekt != null) {
      DataDescription dd = new DataDescription(davConnection.getDataModel().getAttributeGroup(ATG_STATUSMELDUNG), 
          davConnection.getDataModel().getAspect(ASP_STATUSMELDUNG));

      AttributeGroup atg = davConnection.getDataModel().getAttributeGroup(ATG_STATUSMELDUNG);

      Data data = davConnection.createData(atg);

      try {
        long now = (new Date()).getTime();
        data.getItem("Zeitstempel").asTimeValue().setMillis(now);
        data.getItem("Statusmeldung").asTextValue().setText(msg); 
      } catch (NoSuchElementException ex) {
        LOGGER.warning(logPrefix + "Statusmeldung konnte nicht gesendet werden.", ex);
      }

      LOGGER.fine(logPrefix + "Sende Statusmeldung: " + data);

      try {
        davConnection.sendData(new ResultData(statusObjekt, dd, davConnection.getTime(), data));
      } catch (DataNotSubscribedException e) {
        LOGGER.warning(logPrefix + "Statusmeldung konnte nicht gesendet werden.", e);
      } catch (SendSubscriptionNotConfirmed e) {
        try {
          Thread.sleep(MAX_WAIT);
          davConnection.sendData(new ResultData(statusObjekt, dd, davConnection.getTime(), data));
        } catch (Exception e2) {
          LOGGER.warning(logPrefix + "Statusmeldung konnte nicht gesendet werden.", e2);          
        }
      }
    }
  }
}