/**
 * SWE KEx-DatexII (de.bsvrz.kex.datex2)
 * 
 * Copyright (C) 2014 TraffGo Road GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * TraffGo Road GmbH
 * Ostwall 11-19
 * 47798 Krefeld
 * Phone: +49-2151-3295058
 * mailto: info@traffgoroad.de
 */

package de.bsvrz.kex.datex2.plugins.konvertierung.impl;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import javax.naming.ConfigurationException;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.ObjectSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.datex2.allgemein.ObjekteAtgAsp;
import de.bsvrz.kex.datex2.plugins.definition.PluginExport;
import eu.datex2.schema.statisch.measurement.site.messstellen.ConfidentialityValueEnum;
import eu.datex2.schema.statisch.measurement.site.messstellen.CountryEnum;
import eu.datex2.schema.statisch.measurement.site.messstellen.D2LogicalModel;
import eu.datex2.schema.statisch.measurement.site.messstellen.Exchange;
import eu.datex2.schema.statisch.measurement.site.messstellen.HeaderInformation;
import eu.datex2.schema.statisch.measurement.site.messstellen.InformationStatusEnum;
import eu.datex2.schema.statisch.measurement.site.messstellen.InternationalIdentifier;
import eu.datex2.schema.statisch.measurement.site.messstellen.MeasuredOrDerivedDataTypeEnum;
import eu.datex2.schema.statisch.measurement.site.messstellen.MeasurementSiteRecord;
import eu.datex2.schema.statisch.measurement.site.messstellen.MeasurementSiteRecordIndexMeasurementSpecificCharacteristics;
import eu.datex2.schema.statisch.measurement.site.messstellen.MeasurementSiteTable;
import eu.datex2.schema.statisch.measurement.site.messstellen.MeasurementSiteTablePublication;
import eu.datex2.schema.statisch.measurement.site.messstellen.MeasurementSpecificCharacteristics;
import eu.datex2.schema.statisch.measurement.site.messstellen.Point;
import eu.datex2.schema.statisch.measurement.site.messstellen.PointByCoordinates;
import eu.datex2.schema.statisch.measurement.site.messstellen.PointCoordinates;

/**
 * Konvertierung von statischen Daten (Verortung von Umfelddatensensoren) in das DATEX-II Format.
 * 
 * @author TraffGo Road GmbH
 * @version $Id$
 */

public class UDMeasurementSite extends StandardKonvertierungStatisch {
	/**
	 * Typ-Definition fr Umfelddatenmessstellen.
	 */

  private static final String TYP_UFD_MESSSTELLE = "typ.umfeldDatenMessStelle";
	
  /**
   * Mapping von Umfelddatensensoren auf die Umfelddatenmessstelle.
   */
  
	private final Map<SystemObject, SystemObject> ufds2Ufdms = new HashMap<>();

  /**
   * Mapping von Umfelddatenmessstelle auf die Umfelddatensensoren.
   */
  
  private final Map<SystemObject, List<SystemObject> > ufdms2Ufds = new HashMap<>();

  /** Mapping Umfeldsensor -> Index-Nummer. */
  private Map<String, Integer> idxMapping = new HashMap<>();
  
  /***************************************************************************
   * IN DEM FOLGENDEN ABSCHNITT WERDEN STRUKTURELL GLEICHE METHODEN GESAMMELT.
   * 
   * Diese Methoden werden in den einzelnen Plugins wiederholt - allerdings
   * beziehen sich die JAXB-Elemente trotz gleichen Namens dann auf
   * andere Klassen; eine nominale Gleichheit existiert also nicht.
   ***************************************************************************
   */
  /**
   * Gebe den in dieser Klasse verwendeten JAXBContext zurck. 
   * 
   * @throws JAXBException {@see javax.xml.bind.JAXBContext#newInstance(String)}
   * @return JAXBContext Verwendeter Kontext
   */

  @Override
  public JAXBContext getJAXBContext() throws JAXBException {
    if (jaxbContext == null) {
      jaxbContext = JAXBContext.newInstance(D2LogicalModel.class.getPackage().getName());
    }
    return jaxbContext;
  }
  
  /**
   * Setzt Koordinaten aus Data in ein PointCoordinates Objekt.
   * 
   * @param data
   *          Data
   * @return PointCoordinates
   */
  
  private PointCoordinates setzeKoordinaten(final Data data) {
    PointCoordinates pointCoordinates = new PointCoordinates();
    pointCoordinates.setLatitude(WerkzeugSammlung.getLatitude(data));
    pointCoordinates.setLongitude(WerkzeugSammlung.getLongitude(data));
    return pointCoordinates;
  }
  
  /***********************************************************************
   * ENDE DES ABSCHNITTS STRUKTURELL GLEICHER METHODEN
   ***********************************************************************
   */
  
	/**
	 * Konstruktor.
	 * 
	 * @param plugin Referenz auf Plugin-Definitions-Klasse.
	 */
  public UDMeasurementSite(final PluginExport plugin) {
    this.plugin = plugin;
  }

  /**
   * Setzt die allgemeinen Angaben zur Publikation (Exchange, HeaderInformation, Language, PublicationTime).
   * 
   * @param versionFree Wenn true, wird die Version nicht bernommen und nicht persistiert.
   * @param d2l D2LogicalModel
   * @param payload Verwendeter Payload
   */

  private void setzeHeaderInformation(final boolean versionFree, final D2LogicalModel d2l, 
      final MeasurementSiteTablePublication payload) {
    InternationalIdentifier internationalIdentifier = new InternationalIdentifier();
    internationalIdentifier.setCountry(CountryEnum.DE);
    internationalIdentifier.setNationalIdentifier(nationalIdentifier);
    Exchange exchange = new Exchange();
    exchange.setSupplierIdentification(internationalIdentifier);
    d2l.setExchange(exchange);
    HeaderInformation headerInformation = new HeaderInformation();
    headerInformation.setConfidentiality(ConfidentialityValueEnum.NO_RESTRICTION);
    headerInformation.setInformationStatus(InformationStatusEnum.REAL);
    payload.setLang("DE");
   
    payload.setPublicationCreator(internationalIdentifier);
    payload.setHeaderInformation(headerInformation);
    
    payload.setPublicationTime(WerkzeugSammlung.erzeugeVersionsdatum(versionFree));
  }
    
  /**
   * Liefere den in der statischen Publikation verwendeten Index eines Umfelddatensensors zurck.
   * 
   * @param ufdsPid Pid eines Umfelddatensensors
   * @return Verwendeter Index
   */
  public Integer getMeasurementSiteIndex(final String ufdsPid) {
    return idxMapping.get(ufdsPid);
  }
  
  /**
   * Verwendete Abbilung von Umfelddatensensoren (ufds) zu Umfelddatenmessstellen (ufdms).
   * 
   * @return n:1 Abbildung zwischen DaV-Objekte
   */
  public Map<SystemObject, SystemObject> getMappingUfds2Ufdms() {
    return ufds2Ufdms;
  }

//  /**
//   * Verwendete Abbilung von Umfelddatenmessstellen (ufdms) zu Umfelddatensensoren (ufds).
//   * 
//   * @return 1:n Abbildung zwischen DaV-Objekte
//   */
//  public Map<SystemObject, List<SystemObject> > getMappingUfdms2Ufds() {
//    return ufdms2Ufds;
//  }
  
  @Override
  public Object getDatexII(final Collection<ObjekteAtgAsp> objekteAtgAspListe, final CreationOptions... options) {
    boolean versionFree = WerkzeugSammlung.isVersionFree(options);
    String newVersion = WerkzeugSammlung.erzeugeVersionsString(versionFree);

    D2LogicalModel d2l = new D2LogicalModel();
    MeasurementSiteTablePublication payload = new MeasurementSiteTablePublication();
    d2l.setPayloadPublication(payload);
    
    setzeHeaderInformation(versionFree, d2l, payload);
    
    MeasurementSiteTable measurementSiteTable = new MeasurementSiteTable();
    measurementSiteTable.setId(locationGroupId);
    measurementSiteTable.setVersion(newVersion);

    payload.getMeasurementSiteTables().add(measurementSiteTable);
    
    Set<SystemObject> ufdmsSet = new TreeSet<>(ufds2Ufdms.values());
    
    for (SystemObject ufdms : ufdmsSet) {
      MeasurementSiteRecord measurementSiteRecord = new MeasurementSiteRecord();
      
      setzeSiteVersionedId(newVersion, ufdms, measurementSiteRecord);
      setzeSiteLocation(ufdms, measurementSiteRecord);
      setzeSiteCharacteristics(ufdms, measurementSiteRecord);

      if (!measurementSiteRecord.getMeasurementSpecificCharacteristics().isEmpty()) {
        measurementSiteTable.getMeasurementSiteRecords().add(measurementSiteRecord);
      }
    }

    if (istGueltig(measurementSiteTable)) {
      persistiereVersion(versionFree, newVersion);
      return d2l;
    } else {
      return null;
    }
  }

  /**
   * Prft, ob die Tabelle grundstzlich gltig sein kann (hier: ist die Anzahl an SiteRecord-Eintrgen dem Schema 
   * entsprechend ausreichend).
   * 
   * @param measurementSiteTable Zu prfende MeasurementSiteTable
   * @return true = die MeasurementSiteTable kann verwendet werden.
   */
  private boolean istGueltig(final MeasurementSiteTable measurementSiteTable) {
    return measurementSiteTable.getMeasurementSiteRecords().size() >= 2;
  }

  /**
   * Setzt die Id und die Version eines MeasurementSiteRecords mit der Pid einer Umfelddatenmessstelle bzw. der bergebenen
   * Version.
   * 
   * @param newVersion Zu verwendende Version
   * @param ufdms DaV-Objekt Umfelddatenmessstelle
   * @param measurementSiteRecord Der zu setzende SiteRecord
   */
  private void setzeSiteVersionedId(final String newVersion, final SystemObject ufdms, 
      final MeasurementSiteRecord measurementSiteRecord) {
    String pid = ufdms.getPid();
    measurementSiteRecord.setId(pid);
    measurementSiteRecord.setVersion(newVersion);
  }

  /**
   * Setzt die Lokalisierung eines SiteRecords, hier durch Angabe von Geokoordinaten (NetworkLocation - PointCoordinates)
   * aus den Angaben des DaV-Objekts (atg.punktKoordinaten).
   * 
   * @param ufdms DaV-Objekt Umfelddatenmessstelle
   * @param measurementSiteRecord Der zu setzende SiteRecord
   */
  private void setzeSiteLocation(final SystemObject ufdms, final MeasurementSiteRecord measurementSiteRecord) {
    AttributeGroup attributGruppe = ufdms.getDataModel().getAttributeGroup("atg.punktKoordinaten");
    Data data = ufdms.getConfigurationData(attributGruppe);
    Point point = new Point(); // extends NetworkLocation -> extends Location
    PointByCoordinates pointByCoordinates = new PointByCoordinates();
    PointCoordinates pointCoordinates = setzeKoordinaten(data);
    pointByCoordinates.setPointCoordinates(pointCoordinates);
    point.setPointByCoordinates(pointByCoordinates);
    measurementSiteRecord.setMeasurementSiteLocation(point);
  }

  /**
   * Setzt die verfgbaren Informationsarten eines MeasurementSiteRecords.
   * 
   * @param ufdms DaV-Objekt Umfelddatenmessstelle
   * @param measurementSiteRecord Der zu setzende SiteRecord
   */
  private void setzeSiteCharacteristics(final SystemObject ufdms, final MeasurementSiteRecord measurementSiteRecord) {
    int idxCounter = 0;
    Map<MeasuredOrDerivedDataTypeEnum, Integer> dataTypeIndexMapping = new HashMap<>();

    for (SystemObject ufds : ufdms2Ufds.get(ufdms)) {
      MeasurementSpecificCharacteristics msc = new MeasurementSpecificCharacteristics();
      
      switch (ufds.getType().getPid().toLowerCase()) {
      //FahrBahnOberFlchenZustand          
      case "typ.ufdsfahrbahnoberflchenzustand":             
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.ROAD_SURFACE_CONDITION_INFORMATION);
        break;
        //FahrBahnOberFlchenTemperatur     
      case "typ.ufdsfahrbahnoberflchentemperatur":        
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.ROAD_SURFACE_CONDITION_INFORMATION);
        break;
        //GefrierTemperatur      
      case "typ.ufdsgefriertemperatur":                    
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.ROAD_SURFACE_CONDITION_INFORMATION);
        break;
        //RestSalz       
      case "typ.ufdsrestsalz":                     
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.ROAD_SURFACE_CONDITION_INFORMATION);
        break;
        //WasserFilmDicke      
      case "typ.ufdswasserfilmdicke":                    
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.ROAD_SURFACE_CONDITION_INFORMATION);
        break;
        //SchneeHhe       
      case "typ.ufdsschneehhe":                     
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.ROAD_SURFACE_CONDITION_INFORMATION);
        break;
        //LuftTemperatur     
      case "typ.ufdslufttemperatur":                     
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.TEMPERATURE_INFORMATION);
        break;
        //TaupunktTemperatur     
      case "typ.ufdstaupunkttemperatur":                     
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.TEMPERATURE_INFORMATION);
        break;
        //NiederschlagsArt       
      case "typ.ufdsniederschlagsart":                       
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.PRECIPITATION_INFORMATION);
        break;
        //NiederschlagsIntensitt      
      case "typ.ufdsniederschlagsintensitt":                        
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.PRECIPITATION_INFORMATION);
        break;
        //NiederschlagsMenge     
      case "typ.ufdsniederschlagsmenge":                       
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.PRECIPITATION_INFORMATION);
        break;
        //WindGeschwindigkeitMittelWert      
      case "typ.ufdswindgeschwindigkeitmittelwert":                        
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.WIND_INFORMATION);
        break;
        //WindGeschwindigkeitSpitzenWert       
      case "typ.ufdswindgeschwindigkeitspitzenwert":                         
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.WIND_INFORMATION);
        break;
        //WindRichtung       
      case "typ.ufdswindrichtung":                     
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.WIND_INFORMATION);
        break;
        //SichtWeite       
      case "typ.ufdssichtweite":                       
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.VISIBILITY_INFORMATION);
        break;
        //RelativeLuftFeuchte      
      case "typ.ufdsrelativeluftfeuchte":                      
        msc.setSpecificMeasurementValueType(MeasuredOrDerivedDataTypeEnum.HUMIDITY_INFORMATION);
        break;
      default:
        LOGGER.warning(logPrefix + "Sensor-Typ '" + ufds.getType().getPid() 
            + "' nicht untersttzt (pid: " + ufds.getPid() + ").");
        msc.setSpecificMeasurementValueType(null);
        break;
      }
      
      if (msc.getSpecificMeasurementValueType() != null) {
        Integer idx = dataTypeIndexMapping.get(msc.getSpecificMeasurementValueType());
      
        if (idx == null) {
          MeasurementSiteRecordIndexMeasurementSpecificCharacteristics characteristic = 
            new MeasurementSiteRecordIndexMeasurementSpecificCharacteristics();
          characteristic.setMeasurementSpecificCharacteristics(msc);
          idx = Integer.valueOf(idxCounter);
          characteristic.setIndex(idx);
          measurementSiteRecord.getMeasurementSpecificCharacteristics().add(characteristic);
          dataTypeIndexMapping.put(msc.getSpecificMeasurementValueType(), idx);
          idxCounter++;
        }
      
        idxMapping.put(ufds.getPid(), idx);
      }
    }
  }

  /**
   * Hole alle Messstellen zur Zuordnung zu Sensoren.
   * 
   * @return Liste mit SystemObjekten
   */
  
  protected List<SystemObject> getUfdMessstellen() {
    ClientDavInterface davConnection = plugin.getDaVConnection();
    ObjekteAtgAsp objekteAtgAsp = new ObjekteAtgAsp(davConnection, TYP_UFD_MESSSTELLE);
    return objekteAtgAsp.getSystemObjekte();
  }
  
  @Override
  public void initialisiere(final Collection<ObjekteAtgAsp> objects) throws ConfigurationException {
    initialisiereBasisFelder();

    // Stelle Mapping von Sensor zu Messstelle
    
    List<SystemObject> messStellen = getUfdMessstellen();
    
    for (SystemObject messStelle : messStellen) {      
      AttributeGroup atgPunktKoordinaten = plugin.getDaVConnection().getDataModel().getAttributeGroup("atg.punktKoordinaten");
      Data data = messStelle.getConfigurationData(atgPunktKoordinaten);
      if (data != null) {  
        PointCoordinates pointCoordinates = setzeKoordinaten(data);
        
        if (pointCoordinates.getLatitude() > 0 && pointCoordinates.getLongitude() > 0) {
          ObjectSet objectSet = ((ConfigurationObject) messStelle).getObjectSet("UmfeldDatenSensoren");  
          List<SystemObject> ufdsList = objectSet.getElements();
          List<SystemObject> ufdsMap = ufdms2Ufds.get(messStelle);
          if (ufdsMap == null) {
            ufdsMap = new ArrayList<>();
            ufdms2Ufds.put(messStelle, ufdsMap);
          }
          for (SystemObject ufds : ufdsList) {
            ufds2Ufdms.put(ufds, messStelle);
            ufdsMap.add(ufds);
          }
        } 
      }
    }
  }
}
