/**
 * SWE KEx-DatexII (de.bsvrz.kex.datex2)
 * 
 * Copyright (C) 2014 TraffGo Road GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * TraffGo Road GmbH
 * Ostwall 11-19
 * 47798 Krefeld
 * Phone: +49-2151-3295058
 * mailto: info@traffgoroad.de
 */

package de.bsvrz.kex.datex2.plugins.publizierung.impl;

import java.io.BufferedWriter;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;

import de.bsvrz.kex.datex2.plugins.definition.KExD2Plugin;
import de.bsvrz.kex.datex2.plugins.publizierung.PublizierungsMethode;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse zur Speicherung der DATEX-II XML-Struktur in einer Datei.
 * 
 * @author TraffGo Road GmbH
 * @version $Id$
 */

public class XMLFilePublizierung implements PublizierungsMethode {

  /** Zeichencodierung. */
  private static final String CHAR_ENCODING = "UTF-8";
  /** Datex II Namensraum. */
  private static final String DATEX2_NAMESPACE = "http://datex2.eu/schema/2/2_0";

  /** Root-Tag des Datex II Formats. */
  private static final String DATEX2_ROOTTAG = "d2LogicalModel";

  /** subContext gibt an, ob dynamische oder statische Daten bertragen werden. */
  private String subContext = "";

  /** Angabe der SchemaLocation im XML. */
  private String schemaLocation;

  /** Referenz auf Plugin-Definition-Klasse. */
  private KExD2Plugin plugin;

  /** Prfix fr Logging-Ausgaben. */
  private String logPrefix = "";

  /** Logger. */
  private static final Debug LOGGER = Debug.getLogger();

  /** Filename. */
  private String filenameKonfiguration;

  /**
   * Konstruktor. Das bergebene Objekt vom Typ KEx2Plugin ist bei Instantiierung
   * noch nicht vollstndig initialisiert.
   * 
   * @param plugin Referenz auf Plugin-Definitions-Klasse
   * @param subContext Angabe statisch oder dynamisch
   */

  public XMLFilePublizierung(final KExD2Plugin plugin, final String subContext) {
    this.plugin = plugin;
    this.subContext = subContext;
  }  

  /** {@inheritDoc} */
  @Override
  public String getPublizierungsMethodenName() {
    return "Datei";
  }

  /** {@inheritDoc} */
  @Override
  public void publiziere(final Object xml, final JAXBContext context, final String param) throws Exception {
    Marshaller marshaller = context.createMarshaller();
    marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, new Boolean(true));
    marshaller.setProperty(Marshaller.JAXB_ENCODING, CHAR_ENCODING);
    if (schemaLocation != null) {
      marshaller.setProperty(Marshaller.JAXB_SCHEMA_LOCATION, schemaLocation);
    }
    Date zeitstempel = new Date();
    SimpleDateFormat simpleDateFormat = new SimpleDateFormat("yyyyMMdd_HHmmss");
    String fileName = filenameKonfiguration.replaceAll("%D", simpleDateFormat.format(zeitstempel));
    LOGGER.info(logPrefix + "Speichere XML-Datei nach " + fileName);

    Path p = Paths.get(fileName);
    try (BufferedWriter bw = Files.newBufferedWriter(p, Charset.forName(CHAR_ENCODING))) {
      marshaller.marshal(new JAXBElement<Object>(new QName(DATEX2_NAMESPACE, DATEX2_ROOTTAG), Object.class, xml), bw);
    } catch (Exception ex) {
      LOGGER.warning(logPrefix + "Speicherung in Datei " + fileName + " gescheitert.", ex);
    }      
  }

  /** {@inheritDoc} */
  @Override
  public boolean initialisiere() {
    boolean hasAllData = true;
    this.logPrefix = plugin.getKonfigurationsModul().getContext() + ": ";
    try {
      schemaLocation = plugin.getKonfigurationsModul().getData("SchemaLocation-" + subContext, String.class);
    } catch (Exception e) {
      LOGGER.error(logPrefix + "SchemaLocation-" + subContext + " konnte nicht ausgewertet werden.");
      hasAllData = false;
    }
    if (plugin.getKonfigurationsModul().hasData("DateiName-" + subContext)) {
      filenameKonfiguration = plugin.getKonfigurationsModul().getString("DateiName-" + subContext);
    } else {
      hasAllData = false;
    }
    return hasAllData;
  }
}

